"""Entrypoint to run generate a video from a midi file."""

import argparse
from pathlib import Path
from tempfile import TemporaryDirectory

from midi2hands.models.generative import GenerativeHandFormer
from midi2hands.models.onnex.onnex_model import ONNXModel
from midiutils.midi_preprocessor import MidiPreprocessor

from midi2vid.config import Config
from midi2vid.download_soundfont import download_soundfont
from midi2vid.video_generator import VideoGenerator


def main(source_path: Path, video_path: Path, config: Config):
  """Entry point of the program."""
  # check if the sound font is downloaded
  download_soundfont()

  with TemporaryDirectory() as workdir:
    preprocessor = MidiPreprocessor()
    video_generator = VideoGenerator(workdir=Path(workdir), midi_file_path=source_path, config=config)

    events = []
    events = preprocessor.get_midi_events(source_path, max_note_length=int(config.max_note_length))
    if config.estimate_hands:
      model = ONNXModel()
      handformer = GenerativeHandFormer(model=model)
      _, _, y_pred = handformer.inference(events=events, window_size=model.window_size, device="cpu")
      for i, e in enumerate(events):
        e.hand = "left" if y_pred[i] == 0 else "right"
    video_generator.generate_video(events=events, destination_filepath=video_path)


def parse_color(color_str):
  """Convert comma-separated RGB string to list of ints."""
  return [int(x.strip()) for x in color_str.split(",")]


def get_parser():
  parser = argparse.ArgumentParser(description="Configuration for note display")

  # Integer arguments
  parser.add_argument("input", type=str, nargs="?", default=None, help="Input MIDI file path")
  parser.add_argument("output", type=str, nargs="?", default=None, help="Output video file path")

  parser.add_argument("-i", "--input", type=str, dest="input_opt", help="Input MIDI file path (overrides positional)")
  parser.add_argument("-o", "--output", type=str, dest="output_opt", help="Output video file path (overrides positional)")

  parser.add_argument("--max-note-length", type=int, default=80, help="Maximum length of notes")
  parser.add_argument("--n-processors", type=int, default=4, help="Number of processors to use")
  parser.add_argument("--screen-width", type=int, default=1920, help="Screen width in pixels")
  parser.add_argument("--screen-height", type=int, default=1080, help="Screen height in pixels")
  parser.add_argument("--fps", type=int, default=30, help="Frames per second")
  parser.add_argument("--speed", type=int, default=300, help="Note falling speed")
  parser.add_argument("--estimate-hands", action="store_true", default=True, help="Enable hand estimation")
  parser.add_argument("--no-estimate-hands", action="store_false", dest="estimate_hands", help="Disable hand estimation")
  parser.add_argument("--debug", action="store_true", default=False, help="Enable debug mode")

  # Color arguments (RGB triplets)
  parser.add_argument("--white-note-color", type=parse_color, default="255,255,255", help="White note color as RGB (comma-separated)")
  parser.add_argument("--black-note-color", type=parse_color, default="49,49,49", help="Black note color as RGB (comma-separated)")
  parser.add_argument("--background-color", type=parse_color, default="43,43,43", help="Background color as RGB (comma-separated)")
  parser.add_argument("--octave-lines-color", type=parse_color, default="92,92,92", help="Octave lines color as RGB (comma-separated)")
  parser.add_argument("--note-color", type=parse_color, default="179,44,49", help="Note color as RGB (comma-separated)")
  parser.add_argument("--dark-note-color", type=parse_color, default="113,34,36", help="Dark note color as RGB (comma-separated)")
  parser.add_argument("--right-note-color", type=parse_color, default="168,255,145", help="Right hand note color as RGB (comma-separated)")
  parser.add_argument("--left-note-color", type=parse_color, default="176,202,229", help="Left hand note color as RGB (comma-separated)")
  parser.add_argument("--dark-right-note-color", type=parse_color, default="118,208,68", help="Dark right hand note color as RGB (comma-separated)")
  parser.add_argument("--dark-left-note-color", type=parse_color, default="124,142,151", help="Dark left hand note color as RGB (comma-separated)")

  return parser


def commandline_main():
  """Parse the arguments and call main."""
  parser = get_parser()
  args = parser.parse_args()
  # Filter out input/output args and create config with remaining args
  config_args = {k: v for k, v in vars(args).items() if k not in {"input", "output", "input_opt", "output_opt"}}
  config = Config(**config_args)

  parser = argparse.ArgumentParser(description="Convert midi to mp4")

  # Determine final input and output values
  input_file = args.input_opt if args.input_opt is not None else args.input
  output_file = args.output_opt if args.output_opt is not None else args.output

  # Handle missing arguments with sensible defaults or errors
  if input_file is None:
    raise ValueError("No input file specified. Please provide an input MIDI file.")
  if output_file is None:
    # Derive output from input if possible, or use a default
    if input_file is not None:
      output_file = input_file.split(".")[0] + ".mp4"
    else:
      raise ValueError("No output file specified and no input to derive from.")

  output_file = Path(output_file)
  input_file = Path(input_file)

  assert input_file.exists(), f"Input file {input_file} does not exist"

  main(Path(input_file), Path(output_file), config)


if __name__ == "__main__":
  commandline_main()



"""Generates a video of notes from a midi file."""

from __future__ import annotations

import logging
import os
import shutil
import subprocess
import time
from functools import wraps
from multiprocessing import Pool
from pathlib import Path
from typing import Any, Callable, Dict, TypeVar

os.environ["PYGAME_HIDE_SUPPORT_PROMPT"] = "1"
import pygame
from midiutils.types import NoteEvent
from mido import MidiFile  # type: ignore
from pygame import Surface

from midi2vid.config import Config
from midi2vid.note_animation import NoteAnimationModel
from midi2vid.piano import Note, Piano

# Set up logging configuration
logging.basicConfig(level=logging.INFO, format="%(asctime)s - %(levelname)s - %(message)s")

T = TypeVar("T")


def log_performance(func: Callable[..., T]) -> Callable[..., T]:
  @wraps(func)
  def wrapper(*args: Any, **kwargs: Any) -> T:
    start_time = time.perf_counter()
    result = func(*args, **kwargs)
    end_time = time.perf_counter()
    execution_time = (end_time - start_time) * 1000  # Convert to milliseconds
    logging.info(f"{func.__name__} took {execution_time:.4f} ms")
    return result

  return wrapper


def _run_command(command: str, debug: bool = False):
  """Helper function to run shell commands with optional debugging."""
  if debug:
    print(f"Running command: {command}")
  result = subprocess.run(command, shell=True, capture_output=True, text=True)
  if result.returncode != 0:
    print(f"Error executing command:\n{result.stderr}")
    raise RuntimeError(f"Command failed: {command}")
  if debug:
    print(f"Output:\n{result.stdout}")


def get_tempo(mid: MidiFile):
  """returns the tempo in beats per minute"""
  tempo = 500000
  for _ in mid.tracks:
    for msg in mid:
      if msg.type == "set_tempo":
        tempo = msg.tempo
  return 60000000 / tempo


class VideoGenerator:
  """Generates a video from a midi file of the piano notes."""

  def __init__(self, workdir: Path, midi_file_path: Path, config: Config):
    """Will raise an error if the midi_file_path does not exist."""
    self.workdir = workdir
    self.framedir = self.workdir / "frames"
    self.framedir.mkdir()

    self.config = config
    self.midi_file_path = midi_file_path
    self.piano = Piano(screen_width=config.screen_width, screen_height=config.screen_height)
    midi_file = MidiFile(self.midi_file_path)
    self.note_animation_model = NoteAnimationModel(
      bpm=get_tempo(midi_file),
      fps=config.fps,
      ticks_per_beat=midi_file.ticks_per_beat,
      screen_height=config.screen_height,
      note_speed=config.speed,
    )
    self.active_notes: Dict[int, None | NoteEvent] = {i: None for i in self.piano.midi_key_range}
    self.soundfont_path = os.path.join(os.path.dirname(__file__), "data/soundfont.sf2")

  @log_performance
  def _render_frames(self):
    command = (
      f"ffmpeg -framerate {self.config.fps} -i {self.framedir}/%5d.jpg " f"-c:v libx264 -r {self.config.fps} " f"{self.framedir}/video-no-sound.mp4"
    )
    _run_command(command, self.config.debug)

  @log_performance
  def _render_audio(self):
    command = f"fluidsynth -ni {self.soundfont_path} {self.midi_file_path} " f"-F {self.workdir}/audio.wav"
    _run_command(command, self.config.debug)

  @log_performance
  def _merge_audio_and_video(self, destination_filepath: Path):
    command = (
      f"ffmpeg -y -i {self.framedir}/video-no-sound.mp4 "
      f"-i {self.workdir}/audio.wav -c:v copy "
      f"-c:a aac -strict experimental -b:a 192k -f mp4 "
      f"{destination_filepath.absolute()}"
    )
    _run_command(command, self.config.debug)

  def _draw_vertical_lines(self, screen: Surface, screen_height: int):
    left_positions: list[int] = []
    for midi_note_number in self.piano.midi_key_range:
      note = self.piano.get_note(midi_note_number)
      if "c" in note.key or "f" in note.key:
        left = self.piano.get_left_key_pos(note, self.piano.white_key_width, self.piano.black_key_width)
        left_positions.append(left)
    for left_pos in left_positions:
      s = pygame.Surface((1, int(screen_height)))
      s.set_alpha(50)
      s.fill((255, 255, 255))
      screen.blit(s, (left_pos, 0))

  def _draw_piano(self, screen: Surface):
    white_key_width = self.piano.white_key_width
    black_key_width = self.piano.black_key_width
    white_key_height = self.piano.white_key_height
    black_key_height = white_key_height - 0.4 * white_key_height
    top = self.config.screen_height - white_key_height

    self._draw_vertical_lines(screen, self.config.screen_height)

    # Draw white keys
    for midi_note_id in self.piano.midi_key_range:
      note: Note = self.piano.get_note(midi_note_id)
      if "#" in note.key:
        continue
      left_pos = self.piano.get_left_key_pos(note, white_key_width, black_key_width)
      color = self.config.white_note_color
      note_event = self.active_notes[midi_note_id]
      if note_event:
        if note_event.hand == "right":
          color = self.config.right_note_color
        elif note_event.hand == "left":
          color = self.config.left_note_color
        else:
          # if hand is not set, it means it's a note that is played by both
          # hands
          color = self.config.note_color
      rect = pygame.Rect(left_pos, top, white_key_width, white_key_height)
      pygame.draw.rect(screen, color, rect, border_radius=3)
      pygame.draw.rect(screen, (0, 0, 0), rect, 1, border_radius=3)

    # Darw black keys
    for midi_note_id in self.piano.midi_key_range:
      note: Note = self.piano.get_note(midi_note_id)
      if "#" not in note.key:
        continue
      color = self.config.black_note_color
      note_event = self.active_notes[midi_note_id]
      if note_event:
        if note_event.hand == "right":
          color = self.config.dark_right_note_color
        elif note_event.hand == "left":
          color = self.config.dark_left_note_color
        else:
          color = self.config.dark_note_color

      left = self.piano.get_left_key_pos(note, white_key_width, black_key_width)
      rect = pygame.Rect(left, top, black_key_width, black_key_height)
      pygame.draw.rect(screen, color, rect, border_radius=3)

  def _save_frame(self, path: Path, screen: Surface, frame_number: int):
    pygame.image.save(screen, f"{path}/{frame_number:05}.jpg")

  def _is_active(self, note_position: int, duration: float, midi_note_id: int):
    white_key_height_relative_bottom = self.config.screen_height - self.piano.white_key_height
    if note_position > white_key_height_relative_bottom and (note_position - duration) < white_key_height_relative_bottom:
      return True

  def _draw_note(self, note_event: NoteEvent, frame_id: int, screen: pygame.Surface):
    note_padding = 2
    note: Note = self.piano.get_note(note_event.note)
    white_key_width = self.piano.white_key_width
    black_key_width = self.piano.black_key_width

    if "#" in note.key:
      width = self.piano.black_key_width
      if note_event.hand == "right":
        color = self.config.dark_right_note_color
      elif note_event.hand == "left":
        color = self.config.dark_left_note_color
      else:
        color = self.config.dark_note_color
      note_padding = 0
    else:
      width = self.piano.white_key_width - note_padding * 2
      if note_event.hand == "right":
        color = self.config.right_note_color
      elif note_event.hand == "left":
        color = self.config.left_note_color
      else:
        color = self.config.note_color

    # logging.info(f"note {note} is active")
    note_position = round(
      self.note_animation_model.get_note_position(
        start_tick=note_event.start,
        current_frame=frame_id,
        piano_height=int(self.piano.white_key_height),
      )
    )  # the top of the note relative to the top of the screen which is 0

    left_pos = self.piano.get_left_key_pos(note, white_key_width, black_key_width) + note_padding
    # darw the note
    duration = self.note_animation_model.get_note_length(note_event.end - note_event.start)
    note_top = note_position - duration
    white_key_height_relative_bottom = self.config.screen_height - self.piano.white_key_height
    if self._is_active(note_position, duration, note_event.note):
      self.active_notes[note_event.note] = note_event
    # self._set_active_notes(note_position, duration, note_event.note)
    if note_top < white_key_height_relative_bottom:
      pygame.draw.rect(
        screen,
        color,
        pygame.Rect(left_pos, note_top, width, duration),
        border_radius=3,
      )

  def _generate_frame(self, events: list[NoteEvent], frame_id: int, screen: Surface) -> None:
    # reset active notes
    self.active_notes = {note: None for note in self.piano.midi_key_range}
    screen.fill(self.config.background_color)
    # select the events that are in the current frame
    active_events: list[NoteEvent] = self.note_animation_model.get_active_note_events(
      note_events=events,
      current_frame=frame_id,
      screen_height=self.config.screen_height,
      piano_height=int(self.piano.white_key_height),
    )
    # draw the notes that are in the current frame
    for note_event in active_events:  # draw all the notes that will hit a white key
      if "#" not in self.piano.get_note(note_event.note).key:
        self._draw_note(note_event, frame_id, screen)
    for note_event in active_events:  # draw all the notes that will hit a black key
      if "#" in self.piano.get_note(note_event.note).key:
        self._draw_note(note_event, frame_id, screen)
    self._draw_piano(screen)
    self._save_frame(path=self.framedir, screen=screen, frame_number=frame_id)

  def _generate_frame_range(self, events: list[NoteEvent], start_frame: int, end_frame: int):
    screen = pygame.Surface((self.config.screen_width, self.config.screen_height))
    for frame_id in range(start_frame, end_frame):
      self._generate_frame(events, frame_id, screen)

  @log_performance
  def _generate_frames(self, events: list[NoteEvent]):
    total_frames = self.note_animation_model.get_total_number_of_frames(events)
    logging.info(f"Total frames: {total_frames}")
    n_processors = self.config.n_processors
    frames_per_process = total_frames // n_processors
    with Pool(n_processors) as p:
      p.starmap(
        self._generate_frame_range,
        [(events, i * frames_per_process, (i + 1) * frames_per_process) for i in range(n_processors)],
      )

  @log_performance
  def generate_video(
    self,
    events: list[NoteEvent],
    destination_filepath: Path,
  ):
    """Generate a video of a midi file path.

    The events could be generated with the midi_preprocessor.
    """
    logging.info("Generating frames")
    self._generate_frames(events)
    logging.info("Rendering video")
    self._render_frames()
    logging.info("Rendering audio")
    self._render_audio()
    logging.info("Merging audio and video")
    self._merge_audio_and_video(destination_filepath)

    if self.config.debug:
      shutil.copytree(self.workdir, "workdir", dirs_exist_ok=True)



from midiutils.types import NoteEvent
from mido.midifiles import MidiFile


class NoteAnimationModel:
  def __init__(
    self,
    bpm: int,
    fps: int,
    ticks_per_beat: int,
    screen_height: int,
    note_speed: int,
  ):
    self.bpm = bpm
    self.fps = fps
    self.ticks_per_beat = ticks_per_beat
    self.screen_height = screen_height
    self.note_speed = note_speed

    # Derived constants
    self.ticks_per_second = self._calculate_ticks_per_second()
    self.frames_per_tick = self._calculate_frames_per_tick()
    self.pixels_per_frame = self._calculate_pixels_per_frame()
    self.ticks_per_frame = self._calculate_ticks_per_frame()

  def _calculate_ticks_per_second(self) -> float:
    """Calculate how many ticks occur in one second.

    The calculation is based on BPM and ticks per beat.
    """
    beats_per_second = self.bpm / 60  # Convert BPM to BPS(Beats Per Second)
    return beats_per_second * self.ticks_per_beat

  def _calculate_frames_per_tick(self) -> float:
    """Calculate the number of frames that occur for each tick."""
    return 1 / (self.ticks_per_second / self.fps)

  def _calculate_pixels_per_frame(self):
    """Calculate how many pixels a note moves in each frame."""
    return self.note_speed / self.fps

  def _calculate_ticks_per_frame(self):
    return self.ticks_per_second / self.fps

  def get_total_number_of_frames(self, events: list[NoteEvent]):
    """Calculate the total number of frames needed to render all notes."""
    # Find the last note event
    last_event = max(events, key=lambda x: x.end)
    # Calculate the total number of ticks needed to render all notes
    total_ticks = last_event.end
    # Calculate the total number of frames needed to render all notes
    return int(total_ticks / self.ticks_per_frame) + 3 * self.fps

  def get_note_position(
    self, start_tick: int, current_frame: int, piano_height: int = 185
  ):
    """Calculate the position of a note in pixels from the top of the screen 0.
    The position is the bottom of the note, relative to the top of the screen.
    """
    # speed affects the number of pixels a note moves in each frame
    ticks_elapsed = current_frame * self.ticks_per_frame
    ticks_relative_frame = start_tick - ticks_elapsed
    seconds_relative_frame = (
      ticks_relative_frame / self.ticks_per_second
    )  # calculate the hight of the note relative to a frame
    pixels_relative_frame = seconds_relative_frame * self.note_speed
    # the return value is the position where the top left is 0 0.

    note_hight_relavtive_to_bottom = (
      self.screen_height - pixels_relative_frame - piano_height
    )
    return note_hight_relavtive_to_bottom

  def _binary_search(self, note_events: list[NoteEvent], start_tick: float):
    """Find the index of the first note that has not passed the piano."""
    left, right = 0, len(note_events) - 1
    while left <= right:
      mid = left + (right - left) // 2
      if note_events[mid].end < start_tick:
        left = mid + 1
      else:
        right = mid - 1
    return left

  def get_active_note_events(
    self,
    note_events: list[NoteEvent],
    current_frame: int,
    screen_height: int,
    piano_height: int,
  ):
    """Return a list of active notes based on the current frame.

    Active notes are notes that are currently present on the frame.
    Dependent variables:
    - note_events: List of NoteEvent objects, each has a start and end tick.
    - current_frame: The current frame being rendered.
    - screen_height: The height of the screen in pixels.
    - speed: The speed of the notes in pixels per second.
    """
    active_notes: list[NoteEvent] = []
    ticks_elapsed = current_frame * self.ticks_per_frame

    # how many seconds does it take for a note to reach the bottom of the screen
    # how many ticks does it take for a note to reach the bottom of the screen
    # note_speed = pixels per second
    seconds_to_reach_piano_from_top = (
      screen_height - piano_height
    ) / self.note_speed
    ticks_to_reach_piano_from_top = (
      seconds_to_reach_piano_from_top * self.ticks_per_second
    )

    end_ticks = ticks_elapsed + ticks_to_reach_piano_from_top

    # We start from the assumption that notes are sorted by their start tick.
    # find the index of the first note that has not started yet with binary
    # search
    start_index = self._binary_search(note_events, ticks_elapsed - 2000)

    for note in note_events[start_index:]:
      # If the note has not started yet, skip it
      if note.start > end_ticks:
        break
      else:
        active_notes.append(note)
    return active_notes

  def get_total_ticks(self, midi_file: MidiFile):
    return midi_file.ticks_per_beat * (self.bpm / 60) * midi_file.length

  def get_note_length(self, duration: int):
    # Convert duration from ticks to seconds
    # Duration in seconds = (Duration in ticks) / (Ticks per second)
    duration_in_seconds = duration / self.ticks_per_second

    # Convert duration from seconds to pixels
    # Length in pixels = Duration in seconds * Note speed (pixels per second)
    length_in_pixels = duration_in_seconds * self.note_speed

    return length_in_pixels



import os
from pathlib import Path

from gdown import download  # type: ignore


def download_soundfont() -> None:
  # Here I found multiple soundfonts
  # https://sites.google.com/site/soundfonts4u/
  url = "https://drive.google.com/uc?id=1nvTy62-wHGnZ6CKYuPNAiGlKLtWg9Ir9"

  # Define the target directory and file
  data_dir = Path(os.path.dirname(__file__)) / "data"
  data_dir.mkdir(parents=True, exist_ok=True)
  target_file = data_dir / "soundfont.sf2"

  # Download the file
  if not target_file.exists():
    print(f"Downloading file from Google Drive to {target_file}...")
    try:
      download(url, str(target_file), quiet=False)
      print("Download complete.")
    except Exception as e:
      print(f"Failed to download the file: {e}")
      raise e
  else:
    print(f"{target_file} already exists. Skipping download.")


if __name__ == "__main__":
  download_soundfont()



import json
import os
import unittest

from midi2vid.piano import Note, Piano

config_path = os.path.join(os.path.dirname(__file__), "resources/default.json")

with open(config_path, "r") as f:
  config = json.load(f)


class TestPiano(unittest.TestCase):
  def test_get_left_key_pos_start(self):
    # test for left most key that is an a
    note = Note(key="a", octave=0)
    left = Piano().get_left_key_pos(
      note=note, white_key_width=100, black_key_width=60
    )
    assert left == 0

  def test_get_note(self):
    # test for middle c with note number 60
    note = Piano().get_note(midi_note_number=60)
    assert note.key == "c"
    assert note.octave == 4



class Note:
  def __init__(self, key: str, octave: int):
    self.key = key
    self.octave = octave


class Piano:
  def __init__(self, screen_width: int = 1080, screen_height: int = 720):
    self.screen_width = screen_width
    self.screen_height = screen_height
    self.midi_number_to_note = self._generate_piano_midi_representation()
    self.n_white_keys = 52
    self.midi_key_range = range(21, 108 + 1)

    self.white_key_width = self.screen_width / self.n_white_keys
    self.white_key_height = 5 * self.white_key_width
    self.black_key_width = self.white_key_width * 0.6

  def _generate_piano_midi_representation(self) -> dict[int, Note]:
    note_names = [
      "c",
      "c#",
      "d",
      "d#",
      "e",
      "f",
      "f#",
      "g",
      "g#",
      "a",
      "a#",
      "b",
    ]
    start_note = "a"
    note_index = note_names.index(start_note)
    octave = 0
    piano: dict[int, Note] = {}
    for midi_note_number in range(21, 108 + 1):
      piano[midi_note_number] = Note(
        key=f"{note_names[note_index % len(note_names)]}", octave=octave
      )
      note_index += 1
      if note_index % len(note_names) == 0:
        octave += 1
    return piano

  def get_left_key_pos(
    self, note: Note, white_key_width: float, black_key_width: float
  ) -> int:
    start_key = "a"
    octave_distances = {
      "c": 0 * white_key_width,
      "c#": 1 * white_key_width - black_key_width / 2,
      "d": 1 * white_key_width,
      "d#": 2 * white_key_width - black_key_width / 2,
      "e": 2 * white_key_width,
      "f": 3 * white_key_width,
      "f#": 4 * white_key_width - black_key_width / 2,
      "g": 4 * white_key_width,
      "g#": 5 * white_key_width - black_key_width / 2,
      "a": 5 * white_key_width,
      "a#": 6 * white_key_width - black_key_width / 2,
      "b": 6 * white_key_width,
    }

    relative_octave = octave_distances[note.key]
    left = (
      relative_octave
      + (7 * white_key_width * note.octave)
      - octave_distances[start_key]
    )
    return int(left)

  def get_note(self, midi_note_number: int) -> Note:
    return self.midi_number_to_note[midi_note_number]

  def get_key_width(self, note: Note, white_key_width: int) -> float:
    if "#" in note.key:
      return white_key_width * 0.6
    return white_key_width



from dataclasses import dataclass, field
from typing import List


@dataclass
class Config:
  max_note_length: int = 80
  n_processors: int = 4
  screen_width: int = 1920
  screen_height: int = 1080
  fps: int = 30
  speed: int = 300
  white_note_color: List[int] = field(default_factory=lambda: [255, 255, 255])
  black_note_color: List[int] = field(default_factory=lambda: [49, 49, 49])
  background_color: List[int] = field(default_factory=lambda: [43, 43, 43])
  octave_lines_color: List[int] = field(default_factory=lambda: [92, 92, 92])
  note_color: List[int] = field(default_factory=lambda: [179, 44, 49])
  dark_note_color: List[int] = field(default_factory=lambda: [113, 34, 36])
  right_note_color: List[int] = field(default_factory=lambda: [168, 255, 145])
  left_note_color: List[int] = field(default_factory=lambda: [176, 202, 229])
  dark_right_note_color: List[int] = field(default_factory=lambda: [118, 208, 68])
  dark_left_note_color: List[int] = field(default_factory=lambda: [124, 142, 151])
  estimate_hands: bool = True
  debug: bool = False
